{{
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PCM2.1 Analog To Digital Converter Engine
//
// Author: Kwabena W. Agyeman
// Updated: 7/27/2010
// Designed For: P8X32A
// Version: 1.1
//
// Copyright (c) 2010 Kwabena W. Agyeman
// See end of file for terms of use.
//
// Update History:
//
// v1.0 - Original release - 2/28/2010.
// v1.1 - Added variable pin support - 7/27/2010.
//
// For each included copy of this object only one spin interpreter should access it at a time.
//
// Nyamekye,
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Converter Circuit:
//
//                                 3.3V
//                                  |
//                                  C 1NF
//                           100KOHM|
// ADC Left Pin Out Number   ---R---| 0.1UF
//                                  |---C-- Right In
// ADC Left Pin In Number    -------|
//                                  |
//                                  C 1NF
//                                  |
//                                 GND
//
//                                 3.3V
//                                  |
//                                  C 1NF
//                           100KOHM|
// ADC Right Pin Out Number  ---R---| 0.1UF
//                                  |---C-- Right In
// ADC Right Pin In Number   -------|
//                                  |
//                                  C 1NF
//                                  |
//                                 GND
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}}

PUB transferData '' 3 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Helps reliably transfer data to the stereo ADC to record.
'' //
'' // Returns the address of the next data buffer to transfer data to.
'' //
'' // At 8 bits per sample using 1 channel each data buffer is composed of 512 bytes with 512 samples per transfer.
'' //
'' // At 8 bits per sample using 2 channels each data buffer is composed of 512 bytes with 256 samples per transfer.
'' //
'' // At 16 bits per sample using 1 channel each data buffer is composed of 256 words with 256 samples per transfer.
'' //
'' // At 16 bits per sample using 2 channels each data buffer is composed of 256 words with 128 samples per transfer.
'' //
'' // This function waits indefiantly to sync up with the ADC driver.
'' //
'' // Use this function with the longmove/wordmove/bytemove commands and/or a file system to play data.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  repeat while(callerPointer == callePointer)
  result := @dataBlock[256 & callerPointer]
  not callerPointer

PUB clearData '' 3 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Clears all data in all recorder data buffers.
'' //
'' // Helps prevent old data from being recorded when changing to a new data transfer destination.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  bytefill(@dataBlock, 0, 1024)

PUB leftChannelInput '' 3 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the current sample in from the left audio channel. The value returned is a 16 bit signed extended number.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  return leftValue

PUB rightChannelInput '' 3 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the current sample in from the right audio channel. The value returned is a 16 bit signed extended number.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  return rightValue

PUB sampleSignSetup(newSign) '' 4 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Changes whether samples are signed or not.
'' //
'' // NewSign - True to make all samples be saved as signed. False to make all samples be saved as unsigned.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  signedOrUnsigned := newSign

PUB bitsPerSampleSetup(newWidth) '' 7 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Changes the bits per sample, 8 bits or 16 bits.
'' //
'' // NewWidth - New bits per sample to input samples at. Samples are sized by this value. (8 or 16).
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  bitsPerSample := (newWidth => 16)
  computeSampleAdjust

PUB numberOfChannelsSetup(newNumber) '' 4 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Changes the number of channels, 1 channel or 2 channels.
'' //
'' // When set to 1 channel mode the driver will input samples from both channels at once.
'' //
'' // When set to 2 channel mode the driver will input samples from both channels at simultaneously.
'' //
'' // NewNumber - New number of channels to input samples at. Samples are grouped by this value. (1 or 2)
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  numberOfChannels := (newNumber > 1)

PUB sampleRateSetup(newRate) '' 7 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Changes the sample rate. The stereo ADC supportes samples rates from 8HZ to 80KHZ @ 80Mhz.
'' //
'' // NewRate - New sample rate to input samples at. Samples are inputted at the frequency specified by this value.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  sampleRate := (clkfreq / ((newRate <# (clkfreq / constant(80_000_000 / 80_000))) #> 8))
  computeSampleAdjust

PUB ADCEngineStart(leftPinInNumber, leftPinOutNumber, rightPinInNumber, rightPinOutNumber, newRate) '' 15 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Starts up the ADC driver running on a cog.
'' //
'' // Returns true on success and false on failure.
'' //
'' // LeftPinInNumber - Pin used to record left channel audio according to the converter circuit. -1 to disable.
'' // LeftPinOutNumber - Pin used to record left channel audio according to the converter circuit. -1 to disable.
'' // RightPinInNumber - Pin used to record right channel audio according to the converter circuit. -1 to disable.
'' // RightPinOutNumber - Pin used to record right channel audio according to the converter circuit. -1 to disable.
'' // NewRate - New sample rate to input samples at. Samples are inputted at the frequency specified by this value.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  ADCEngineStop
  if(chipver == 1)

    sampleRateSetup(newRate)

    newRate := ((leftPinOutNumber <# 31) #> 0)
    leftCounterSetup := ( (constant(%0_1001 << 26) + (newRate << 9) + ((leftPinInNumber <# 31) #> 0)) {
                        } & (leftPinInNumber <> -1) )

    leftPinInNumber := ((rightPinOutNumber <# 31) #> 0)
    rightCounterSetup := ( (constant(%0_1001 << 26) + (leftPinInNumber << 9) + ((rightPinInNumber <# 31) #> 0)) {
                         } & (rightPinInNumber <> -1) )

    outputMask := (((leftPinOutNumber <> -1) & (|<newRate)) | ((rightPinOutNumber <> -1) & (|<leftPinInNumber)))

    leftValueAddress := @leftValue
    rightValueAddress := @rightValue
    multiplyOrDivideAddress := @multiplyOrDivide
    unsignedOrSignedAddress := @signedOrUnsigned
    bitsPerSampleAddress := @bitsPerSample
    numberOfChannelsAddress := @numberOfChannels
    dataBlockAddress := @dataBlock
    sampleAdjustAddress := @sampleAdjust
    callePointerAddress := @callePointer

    cogNumber := cognew(@initialization, @sampleRate)
    result or= ++cogNumber

PUB ADCEngineStop '' 3 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Shuts down the ADC driver running on a cog.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if(cogNumber)
    cogstop(-1 + cogNumber~)

PRI computeSampleAdjust ' 3 Stack Longs

  result := ($FF_FF >> (8 & (not(bitsPerSample))))

  if(sampleRate =< result)

    sampleAdjust := (result / sampleRate)
    multiplyOrDivide := false

  else

    sampleAdjust := ((sampleRate / result) + 1)
    multiplyOrDivide := true

DAT

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'                       ADC Driver
' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

                        org     0

' //////////////////////Initialization/////////////////////////////////////////////////////////////////////////////////////////

initialization          mov     ctra,            leftCounterSetup           ' Setup counter I/O pins.
                        mov     ctrb,            rightCounterSetup          '
                        mov     frqa,            #1                         '
                        mov     frqb,            #1                         '
                        mov     dira,            outputMask                 '
                        mov     recorderPointer, dataBlockAddress           '

                        rdlong  timeCounter,     par                        ' Setup timing.
                        add     timeCounter,     cnt                        '

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'                       Recorder
' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

outerLoop               rdbyte  buffer,          numberOfChannelsAddress wz ' Setup recorder mode.
                        muxz    recorderMode,    #$1                        '
                        rdbyte  buffer,          bitsPerSampleAddress wz    '
                        muxz    recorderMode,    #$2                        '

                        test    recorderMode,    #$1 wc                     ' Setup counter.
                        mov     counter,         #128                       '
if_c_or_z               add     counter,         #128                       '
if_c_and_z              add     counter,         #256                       '

' //////////////////////Inner Loop/////////////////////////////////////////////////////////////////////////////////////////////

innerLoop               rdlong  buffer,          par                        ' Wait until next sample input period.
                        waitcnt timeCounter,     buffer                     '

                        mov     sampleBuffer,    phsa                       ' Input and set value.
                        sub     sampleBuffer,    leftDelta                  '
                        add     leftDelta,       sampleBuffer               '
                        add     leftAverage,     sampleBuffer               '
                        shr     leftAverage,     #1                         '
                        mov     sampleBuffer,    leftAverage                '
                        movs    encodedSample,   #leftValueAddress          '
                        call    #encode                                     '

                        mov     sampleBuffer,    phsb                       ' Input and set value.
                        sub     sampleBuffer,    rightDelta                 '
                        add     rightDelta,      sampleBuffer               '
                        add     rightAverage,    sampleBuffer               '
                        shr     rightAverage,    #1                         '
                        mov     sampleBuffer,    rightAverage               '
                        movs    encodedSample,   #rightValueAddress         '
                        test    recorderMode,    #1 wc                      '
if_c                    wrlong  buffer,          rightValueAddress          '
if_nc                   call    #encode                                     '

                        djnz    counter,         #innerLoop                 ' Loop.

' //////////////////////Outer Loop/////////////////////////////////////////////////////////////////////////////////////////////

                        rdword  buffer,          callePointerAddress wz     ' Switch data block pointer.
                        sumz    buffer,          #1                         '
                        wrword  buffer,          callePointerAddress        '
if_nz                   mov     recorderPointer, dataBlockAddress           '

                        jmp     #outerLoop                                  ' Loop.

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'                       Encode Value
' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

encode                  rdbyte  buffer,          multiplyOrDivideAddress wz ' Setup to divide or multiply.
                        rdlong  mathArgument,    sampleAdjustAddress        '
                        mov     mathResult,      #0                         '
                        mov     buffer,          #16                        '

multiplyLoop if_z       shr     sampleBuffer,    #1 wc                      ' Preform multiplication.
if_z_and_c              add     mathResult,      mathArgument               '
if_z                    shl     mathArgument,    #1                         '
if_z                    tjnz    sampleBuffer,    #multiplyLoop              '

if_nz                   shl     mathArgument,    #15                        ' Preform division.
divideLoop if_nz        cmpsub  sampleBuffer,    mathArgument wc            '
if_nz                   rcl     sampleBuffer,    #1                         '
if_nz                   rcl     mathResult,      #1                         '
if_nz                   djnz    buffer,          #divideLoop                '

                        test    recorderMode,    #$2 wc                     ' Modify data depending on sign and size.
                        rdbyte  buffer,          unsignedOrSignedAddress wz '
if_c                    max     mathResult,      #$FF                       '
if_nc                   max     mathResult,      wordMax                    '

                        mov     buffer,          mathResult                 ' Update main memory sample long.
if_c                    shl     buffer,          #8                         '
                        sub     buffer,          wordAdjust                 '
encodedSample           wrlong  buffer,          0                          '

if_nz_and_c             sub     mathResult,      #$80                       ' Adjust signedness.
if_nz_and_nc            sub     mathResult,      wordAdjust                 '

if_c                    wrbyte  mathResult,      recorderPointer            ' Write data depending on size.
if_c                    add     recorderPointer, #1                         '
if_nc                   wrword  mathResult,      recorderPointer            '
if_nc                   add     recorderPointer, #2                         '

encode_ret              ret                                                 ' Return.

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'                       Data
' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

leftDelta               long    0                                           ' Delta initialized to zero.
rightDelta              long    0                                           ' Delta initialized to zero.
leftAverage             long    0                                           ' Moving left average.
rightAverage            long    0                                           ' Moving right average.
wordMax                 long    $FF_FF                                      ' Edits word unsigned value.
wordAdjust              long    $80_00                                      ' Edits word signed value.

' //////////////////////Configuration Settings/////////////////////////////////////////////////////////////////////////////////

leftCounterSetup        long    0
rightCounterSetup       long    0
outputMask              long    0

' //////////////////////Addresses//////////////////////////////////////////////////////////////////////////////////////////////

leftValueAddress        long    0
rightValueAddress       long    0
multiplyOrDivideAddress long    0
unsignedOrSignedAddress long    0
bitsPerSampleAddress    long    0
numberOfChannelsAddress long    0
dataBlockAddress        long    0
sampleAdjustAddress     long    0
callePointerAddress     long    0

' //////////////////////Run Time Variables/////////////////////////////////////////////////////////////////////////////////////

buffer                  res     1
counter                 res     1
recorderPointer         res     1
recorderMode            res     1
sampleBuffer            res     1
timeCounter             res     1
mathResult              res     1
mathArgument            res     1

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

                        fit     496

DAT

' //////////////////////Global Variable Array//////////////////////////////////////////////////////////////////////////////////

sampleRate              long    0                                              ' Driver sample rate.

leftValue               long    0                                              ' Driver left channel PSHA value.
rightValue              long    0                                              ' Driver right channel PSHB value.

multiplyOrDivide        byte    0                                              ' Flag for multipling or dividing the sample.
signedOrUnsigned        byte    0                                              ' Flag for signed or unsigned samples.
bitsPerSample           byte    0                                              ' Flag for 8 or 16 bit samples.
numberOfChannels        byte    0                                              ' Flag for 1 or 2 channels.

dataBlock               word    0[512]                                         ' Double recordback buffer.

sampleAdjust            long    0                                              ' Multiply or divide by number.

callerPointer           word    0                                              ' Mask for switching between recordback buffers.
callePointer            word    0                                              ' Mask for switching between recordback buffers.

cogNumber               byte    0                                              ' Cog ID.

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

{{

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                  TERMS OF USE: MIT License
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
// Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}}